#!/usr/bin/env python

'''Process manager for Blue Gene/L systems'''
__revision__ = '$Revision: 1.14 $'

import atexit
from os import dup2, execl, fork, kill, setgid, setuid, system, waitpid, WNOHANG
from pwd import getpwnam
import signal
import sys
from sys import argv
from syslog import syslog, LOG_INFO, LOG_ERR
from tempfile import mktemp
from time import time
import os
from ConfigParser import ConfigParser

from elementtree.ElementTree import Element, SubElement

from sss.server import Server, AddEvent
from sss.restriction import Data, DataSet, ID


class ProcessGroupCreationError(Exception):
    '''ProcessGroupCreation Error is used when not enough information is specified'''
    pass

class ProcessGroup(Data):
    '''The ProcessGroup class implements all stages of running parallel processes'''
    # read in config from cobalt.conf
    _configfields = ['mmcs_server_ip', 'db2_instance', 'bridge_config', 'mpirun', 'db2_properties', 'db2_connect']
    _config = ConfigParser()
    if '-C' in argv:
        _config.read(argv[argv.index('-C') + 1])
    else:
        _config.read('/etc/cobalt.conf')
    if not _config._sections.has_key('bgpm'):
        print '''"bgpm" section missing from cobalt config file'''
        raise SystemExit, 1
    config = _config._sections['bgpm']
    mfields = [field for field in _configfields if not config.has_key(field)]
    if mfields:
        print "Missing option(s) in cobalt config file: %s" % (" ".join(mfields))
        raise SystemExit, 1

    def __init__(self, element, pgid):
        Data.__init__(self, element)
        self.element.tag = 'process-group'
        self.element.set('pgid', pgid)
        self.element.set('state', 'initializing')
        if self.element.get('outputfile', False):
            self.outlog = self.element.get('outputfile')
        else:
            self.outlog = mktemp()            
        if self.element.get('errorfile', False):
            self.errlog = self.element.get('errorfile')
        else:
            self.errlog = mktemp()
        self.pid = fork()
        if self.element.find('.//host-spec') != None:
            partition = self.element.find('.//host-spec').text
        else:
            raise ProcessGroupCreationError, "partition"
        try:
            userid, groupid = getpwnam(self.element.get('submitter'))[2:4]
        except KeyError:
            raise ProcessGroupCreationError, "user/group"
        
        program = self.element.findall(".//process-spec")[0].get('exec')
        cwd = self.element.findall(".//process-spec")[0].get('cwd')
        pnum = self.element.get('totalprocs')
        mode = self.element.get('mode', 'co')
        args = " ".join([entry.get('value') for entry in self.element.findall(".//arg")])
        envs = " ".join(["%s=%s" % (entry.get('name'), entry.get('value')) for entry in
                         self.element.findall(".//env")])

        if not self.pid:
            atexit._atexit = []
            try:
                setgid(groupid)
                setuid(userid)
            except OSError:
                syslog(LOG_ERR, "Failed to change userid/groupid for PG %s" % (self.element.get("pgid")))
                sys.exit(0)
            #system("/bgl/BlueLight/ppcfloor/bglsys/bin/db2profile > /dev/null 2>&1")
            system("%s > /dev/null 2>&1" % (self.config['db2_connect']))
            os.environ["DB_PROPERTY"] = self.config['db2_properties']
            os.environ["BRIDGE_CONFIG_FILE"] = self.config['bridge_config']
            os.environ["MMCS_SERVER_IP"] = self.config['mmcs_server_ip']
            os.environ["DB2INSTANCE"] = self.config['db2_instance']
            os.environ["LD_LIBRARY_PATH"] = "/u/bgdb2cli/sqllib/lib"
            try:
                err = open(self.errlog, 'w')
                dup2(err.fileno(), sys.__stderr__.fileno())
            except IOError:
                syslog(LOG_ERR, "Failed to open stderr file %s. Stderr will be lost" % (self.errlog))
            try:
                out = open(self.outlog, 'w')
                dup2(out.fileno(), sys.__stdout__.fileno())
            except IOError:
                syslog(LOG_ERR, "Failed to open stdout file %s. Stdout will be lost" % (self.outlog))
            null = open('/dev/null', 'r')
            dup2(null.fileno(), sys.__stdin__.fileno())
            cmd = (self.config['mpirun'], "mpirun", '-np', pnum, '-partition', partition,
                               '-mode', mode, '-cwd', cwd, '-exe', program)
            if args != '':
                cmd = cmd + ('-args', args)
            if envs != '':
                cmd = cmd + ('-env', envs)
            syslog(LOG_INFO, "Running %s" % (" ".join(cmd)))
            apply(execl, cmd)
            sys.exit(0)
        else:
            self.element.set('state', 'running')
            syslog(LOG_INFO, "ProcessGroup %s Started on partition %s. pid: %s" % (pgid,
                                                                                   partition, self.pid))
            AddEvent("process-manager", "process_start", pgid)

    def FinishProcess(self, status):
        '''Handle cleanup for exited process'''
        # process has already been waited on
        self.element.attrib['state'] = 'finished'
        syslog(LOG_INFO, "ProcessGroup %s Finshed. pid %s" % (self.element.get('pgid'), self.pid))
        AddEvent("process-manager", "process_end", self.element.get('pgid'))
        output = SubElement(self.element, "output")
        error = SubElement(self.element, "error")
        # split out rank additions
        #output.text = "".join([line.split(':', 1)[1][1:] for line in open(self.outlog).readlines()])
        #error.text = "\n".join([line.split(':', 1)[1] for line in open(self.errlog).readlines()])
        if not self.element.get('outputfile', False):
            output.text = open(self.outlog).read()
        if not self.element.get('errorfile', False):
            error.text = open(self.errlog).read()
        estatus = SubElement(self.element, "exit-status")
        SubElement(estatus, "exit-code", status=str(status), host='BG/L')

    def Signal(self, signame):
        '''Send a signal to a process group'''
        try:
            kill(self.pid, getattr(signal, signame))
        except OSError, error:
            syslog(LOG_ERR, "Signal failure for pgid %s:%s" % self.element.get('pgid'), error.strerror)

    def Kill(self):
        '''Kill Blue Gene job. This method is more vicious; it is processed through the bridge API
        Not Yet Implemented'''
        pass

class BGProcessManager(Server):
    '''The BGProcessManager supports the BG/L process execution model'''
    __implementation__ = 'bgpm'
    __component__ = 'process-manager'
    __dispatch__ = {'create-process-group':'AddPG_cb',
                    'signal-process-group':'XSigPG',
                    'get-process-group-info':'pg.Get',
                    'wait-process-group':'pg.Del',
                    'signal-process-group':'XSigPG',
                    'kill-process-group':'XKillPG'}
    __validate__ = 0

    def __setup__(self):
        self.pg = DataSet("process-groups", "process-group", ProcessGroup, ID(), True)
        self.ignore = []
        self.lastwait = 0

    def __progress__(self):
        if (time() - self.lastwait) > 60:
            while True:
                try:
                    self.lastwait = time()
                    (pid, stat) = waitpid(-1, WNOHANG)
                except OSError:
                    break
                if pid == 0:
                    break
                pgrps = [pgrp for pgrp in self.pg if pgrp.pid == pid]
                if len(pgrps) == 0:
                    syslog(LOG_ERR, "Failed to locate process group for pid %s" % (pid))
                elif len(pgrps) == 1:
                    pgrps[0].FinishProcess(stat)
                else:
                    syslog(LOG_ERR, "Got more than one match for pid %s" % (pid))
        return 0

    def XSigPG(self, xml, (peer, port)):
        return self.pg.Get(xml, (peer, port), self.SigPG)
    
    def SigPG(self, pgrp, args):
        self.ignore.append(pgrp.Signal(args['signal'], args.get('scope', 'single')))

    def XKillPG(self, xml, (peer, port)):
        return self.pg.Get(xml, (peer, port), self.KillPG)

    def KillPG(self, pgrp, args):
        self.ignore.append(pgrp.Signal('SIGINT'))

    def AddPG_cb(self, xml, portinfo):
        '''This is needed since the pm syntax isnt quite restriction-like'''
        try:
            pgrp = ProcessGroup(xml, str(self.pg.idalloc.get()))
            self.pg.data.append(pgrp)
            return pgrp.Fetch(xml)
        except ProcessGroupCreationError, emsg:
            syslog(LOG_ERR, "Failed to create process group: %s" % emsg)
            return Element('error', type='Semantic', reason='Incomplete Specification')
    
    def SigChildHand(self, sig, frame):
        '''Dont Handle SIGCHLDs'''
        pass
    
if __name__ == '__main__':
    from getopt import getopt, GetoptError
    try:
        (opts, arg) = getopt(argv[1:], 'dC:', ['daemon='])
    except GetoptError,msg:
        print "%s\nUsage:\nbgpm.py [-d] [-C config file] [--daemon <pidfile>]" % (msg)
        raise SystemExit, 1
    daemon = [item for item in opts if item[0] == '--daemon']
    ldebug = len([item for item in opts if item[0] == '-d'])
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0][1])
    s = BGProcessManager(debug=ldebug)
    s.ServeForever()
