#!/usr/bin/env python
# $Id: cam.py 1.7 05/06/29 15:07:09-05:00 zach@fluffhead.mcs.anl.gov $

from elementtree.ElementTree import Element, SubElement, parse, tostring
from sss.restriction import Data, DataSet
from sss.server import Server
from validate import validate, ValidationException
from syslog import syslog, LOG_ERR, LOG_INFO, LOG_DEBUG
from random import random
from sys import argv
from logging import getLogger, FileHandler, Formatter, INFO, WARNING, DEBUG
from ConfigParser import ConfigParser

class AmountError(Exception):
    pass

class Project(Data):
    def AddUser(self, user):
        if not [x for x in self.element.findall('user') if x.attrib['name'] == user]:
            self.element.append(Element('user', name=user))

    def DelUser(self, user):
        u = [x for x in self.element.findall('user') if x.attrib['name'] == user]
        if len(u) == 1:
            self.element.remove(u[0])

    def AddLien(self, amount):
        amount = float(amount)
        nodeseconds = float(self.element.get('allocation',default='-1'))#if there is no allocation, no limit 
        
        if (amount <= nodeseconds):
            nodeseconds -= amount
        elif nodeseconds != -1:
            raise AmountError

        id = random()
        u = [x for x in self.element.findall('lien') if x.attrib['id'] == id]
        while len(u) != 0:
            id = random()
            u = [x for x in self.element.findall('lien') if x.attrib['id'] == id]

        self.element.set('allocation',str(nodeseconds))    
        self.element.append(Element('lien',id=str(id),amount=str(amount)))
        return id

    def DelLien(self, id):
        u = [x for x in self.element.findall('lien') if x.attrib['id'] == id]
        nodeseconds = float(self.element.get('allocation')) + float(u[0].get('amount'))
        self.element.set('allocation',str(nodeseconds))
        self.element.remove(u[0])
        
    def CommitLien(self, id, actual):
        u = [x for x in self.element.findall('lien') if x.attrib['id'] == id]
        d = ( float(u[0].get('amount')) - float(actual) )
        nodeseconds = float(self.element.get('allocation'))
        if (nodeseconds > -1):
            nodeseconds = float( self.element.get('allocation'))+d
        else:
            nodeseconds = -1
        self.element.set('allocation',str(nodeseconds))
        self.element.remove(u[0])

    def SetProject(self, allocation):
        self.element.set('allocation',str(allocation))

class allocmgr(Server):
    __component__ = 'allocation-manager'
    __implementation__ = 'cam'
    __statefields__ = ['projects']
    __dispatch__ = {'add-project':'projects.Add', 'get-project':'projects.Get', 'del-project':'projects.Del', 'set-project':'XSetProject', 'add-user':'XAddUser','del-user':'XDelUser','add-lien':'XAddLien','del-lien':'XDelLien','commit-lien':'XCommit'}
    __validate__ = 1

    def __setup__(self):
        _configfields = ['log_dir']      
        _config = ConfigParser()
        if '-C' in argv:
            _config.read(argv[argv.index['C'] + 1])
        else:
            _config.read('/etc/cobalt.conf')
        if not _config._sections.has_key('cam'):
            print '''"cam" section missing from cobalt config file'''
            raise SystemExit, 1
        self.config = _config._sections['cam']
        mfields = [field for field in _configfields if not self.config.has_key(field)]
        if mfields:
            print "Missing option(s) in cobalt config file: %s" % (" ".join(mfields))
            raise SystemExit, 1

        self.projects=DataSet("projects","project",Project,False,False)
        self.logger = getLogger('cqm')
        self.hdlr = FileHandler('%s/cam-%s.log' % (self.config['log_dir'], strftime("%m-%d-%y", localtime())))
        self.formatter = Formatter('%(asctime)s;%(message)s')
        self.hdlr.setFormatter(self.formatter)
        self.logger.addHandler(self.hdlr)
        self.logger.setLevel(INFO)
        self.prevdate = strftime("%m-%d-%y", localtime())

    def __progress__(self):
        newdate = strftime("%m-%d-%y", localtime())
        if newdate != self.prevdate:
            self.logger.info('ending logfile')
            self.logger.removeHandler(self.hdlr)
            self.hdlr.close()
            self.hdlr = FileHandler('%s/cam-%s.log' % (self.config['log_dir'], strftime("%m-%d-%y", localtime())))
            self.hdlr.setFormatter(self.formatter)
            self.logger.addHandler(self.hdlr)
            self.logger.info('starting logfile')
            self.prevdate = strftime("%m-%d-%y", localtime())
        return 1
        
    def XAddUser(self, xml, (peer,port)):
        return self.projects.Get(xml, (peer,port), self.AddUser)

    def AddUser(self, proj, args):
        proj.AddUser(args['user'])

    def XDelUser(self, xml, (peer,port)):
        return self.projects.Get(xml, (peer,port), self.DelUser)

    def DelUser(self, proj, args):
        proj.DelUser(args['user'])

    def XAddLien(self, xml, (peer,port)):
        result = Element('projects')
        for s in xml.findall('project'):
            for item in [x for x in self.projects.data if x.Match(s)]:
                id =  self.AddLien(item, xml.attrib)
                result.append(item.Fetch(s))
        if id:result.append(Element('lien',id=str(id)))
        return result

    def AddLien(self, proj, args):
        try:
            id = proj.AddLien(args['amount'])
        except AmountError, e:
            syslog(LOG_ERR, "Failed to allocate enough time for project: requested %s, only %s available" % (args['amount'], proj.element.get('allocation')))
            return
        if id:
            self.logger.info('added lien of %s to project %s' % (args['amount'], proj.element.get('name')))
            self.logger.info('queued;%s' % (tostring(proj.element)))
            return id

    def XSetProject(self, xml, (peer,port)):
        return self.projects.Get(xml, (peer, port), self.SetProject)

    def SetProject(self, proj, args):
        self.logger.info('changed allocation from %s to %s for project %s' % (proj.element.get('name'), args['allocation'], proj.element.get('name')))
        proj.SetProject(args['allocation'])

    def XDelLien(self, xml, (peer,port)):
        return self.projects.Get(xml, (peer,port), self.DelLien)

    def DelLien(self, proj, args):
        proj.DelLien(args['id'])
        self.logger.info('deleted lien id %s from project %s' % (args['id'], proj.element.get('name')))

    def XCommit(self, xml, (peer,port)):
        return self.projects.Get(xml, (peer,port), self.CommitLien)

    def CommitLien(self, proj, args):
        proj.CommitLien(args['id'],args['actual'])
        self.logger.info('committed lien for project %s; actual time %s' % (proj.element.get('name'), args['actual']))

if __name__ == '__main__':
    from getopt import getopt, GetoptError
    from sys import argv,exit
    from time import localtime, strftime

    try:
        (opts,args)=getopt(argv[1:],'d',['daemon='])
    except GetoptError,msg:
        print "%s\nUsage:\ncam.py [-d] [--daemon <pidfile>]"%(msg)
        exit(1)
    daemon=[x[1] for x in opts if x[0] == '--daemon']
    debug=len([x for x in opts if x[0] == '-d'])
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0])
    a = allocmgr(debug=debug)
    a.ServeForever()
