#!/usr/bin/env python
# $Id: cqm.py 1.25 05/07/19 10:16:12-05:00 desai@topaz.mcs.anl.gov $

'''Cobalt Queue Manager'''
__revision__ = '$Revision: 1.25 $'

from elementtree.ElementTree import Element, SubElement, XML, tostring
from xml.sax.saxutils import escape
from os import readlink, symlink, system, unlink
from signal import signal, SIGCHLD, SIG_DFL
from sys import argv, exc_info
from syslog import syslog, LOG_INFO, LOG_ERR, LOG_DEBUG
from time import time
from traceback import extract_tb
#from smtplib import SMTP
from ConfigParser import ConfigParser

from sss.restriction import Data, DataSet, ID
from sss.server import EventReceiver, AddEvent
from sss.ssslib import comm_lib, ConnectError, CommunicationError, ProtocolError, SocketError, LookupError as SSSLookupError

class ProcessManagerError(Exception):
    '''This error occurs when communications with the process manager fail'''
    pass

class TimerException(Exception):
    '''This error occurs when timer methods are called in the wrong order'''
    pass

class Timer(object):
    def __init__(self):
        pass
    
    def Start(self):
        self.start = time()
        
    def Stop(self):
        self.stop = time()
        
    def Check(self):
        try:
            return self.stop - self.start
        except:
            raise TimerException, "StateError"

    def Elapsed(self):
        try:
            return time() - self.start
        except:
            raise TimerException, "StateError"

comm = comm_lib(debug=0)

class Job(Data):
    '''The Job class is an object corresponding to the qm notion of a queued job, including steps'''
    def __init__(self, e, jobid):
        self.element = e
        self.attr = self.element.attrib
        self.attr['jobid'] = str(jobid)
        self.attr['state'] = 'queued'
        self.attr['attribute'] = 'compute'
        self.element.set('location', 'none')
        self.staged = 0
        self.killed = False
        self.timers = {}
        self.timers['queue'] = Timer()
        self.timers['queue'].Start()
        self.timers['/usr/sbin/prologue'] = Timer()
        self.timers['user'] = Timer()
        self.timers['/usr/sbin/epilogue'] = Timer()
        self.pgid = {}
        self.spgid = {}
        self.steps = ['StageInit', 'FinishStage', 'RunPrologue',
                      'RunUserJob', 'RunEpilogue', 'FinishUserPgrp', 'FinalizeStage', 'Finish']
        #self.steps=['StageInit','FinishStage','RunPrologue','RunUserJob','RunEpilogue','FinalizeStage','Finish']
        self.stageid = None
        if not self.attr.get('type', None):
            self.attr['type'] = 'mpish'
        AddEvent("queue-manager", "job-submitted", self.attr['jobid'])
        self.SetActive()

    def WriteOutput(self):
        try:
            open("%s/%s.output" % (self.attr['outputdir'], self.attr['jobid']), 'w').write(self.output)
            open("%s/%s.error" % (self.attr['outputdir'], self.attr['jobid']), 'w').write(self.error)
        except IOError:
            syslog(LOG_ERR, "Failed to write to userdir for job %s writing to /tmp instead" % self.attr['jobid'])
            open("%s/%s.output" % ('/tmp', self.attr['jobid']), 'w').write(self.output)
            open("%s/%s.error" % ('/tmp', self.attr['jobid']), 'w').write(self.error)

    def CheckProject(self):
        global comm
        try:
            req = '''<get-project><project name='%s'><user name='%s'/></project></get-project>''' % (
                self.attr['project'], self.attr['user'])
            h = comm.ClientInit('allocation-manager')
            comm.SendMessage(h, req)
            r = comm.RecvMessage(h)
            comm.ClientClose(h)
        except:
            syslog(LOG_ERR, "Failed to contact allocation manager")
            print "failed to contact AM"
            return
        
        if not len(XML(r).getchildren()):
            # they don't have a project
            syslog(LOG_ERR, "Allocation error for user %s submitting with project %s (project doesn't exist)"%(self.attr['user'], self.attr['project']))
            self.attr['state'] = 'am-error'
            self.steps = []
            return
        else:
            try:
                #add-lien is specified in time of seconds
                req = '''<add-lien amount='%s'><project name='%s'/></add-lien>''' % (
                    str(float((self.attr['walltime'])) * 60), self.attr['project'])

                h = comm.ClientInit('allocation-manager')
                comm.SendMessage(h, req)
                r = comm.RecvMessage(h)
                comm.ClientClose(h)
            except:
                syslog(LOG_ERR, "Failed to contact allocation manager")
                self.attr['state'] = 'am-error'
                self.steps = []
                return

            try:
                self.lienID = XML(r).find('lien').get('id')
            except:
                syslog(LOG_ERR, "Allocation error for user %s submitting with project %s (insufficient balance)"%(self.attr['user'], self.attr['project']))
                self.attr['state'] = 'am-error'
                self.steps = []
                return
        return

    def SetActive(self):
        self.active = True

    def SetPassive(self):
        self.active = False
        
    def HasPG(self, pgid):
        if pgid in self.pgid.values():
            return 1
        return 0

    def Finish(self):
        #am-stuff
        global comm
        try:
            time = int(self.timers['user'].Check()) * len(self.element.get('location').split(':'))
            req = '''<commit-lien id='%s' actual='%s'><project name='%s'/></commit-lien>''' % (str(self.attr['lienID']), str(time), self.attr['project'])
            h = comm.ClientInit('allocation-manager')
            comm.SendMessage(h, req)
            r = comm.RecvMessage(h)
            comm.ClientClose(h)
        except:
            pass
                                                                
        self.attr['state'] = 'done'
        self.SetPassive()
        AddEvent("queue-manager", "job-completed", self.attr['jobid'])

    def GetScript(self):
        return self.element.find('script').text

    def Progress(self):
        if not self.steps:
            syslog(LOG_ERR, "Manually setting passive for job %s" % (self.attr['jobid']))
            self.SetPassive()
            return
        syslog(LOG_INFO, "running step %s for job %s" % (self.steps[0], self.attr['jobid']))
        getattr(self, self.steps[0])()
        try:
            self.steps = self.steps[1:]
        except IndexError:
            syslog(LOG_ERR, "No more job steps for job %s" % (self.attr['jobid']))
            self.attr['state'] = 'error'
        except:
            syslog(LOG_ERR, "Unexpected failure jobid:%s step:%s" % (self.attr['jobid'], self.steps[0]))
            (t, v, tb) = exc_info()
            syslog(LOG_ERR, "Error: %s: %s" % (t, v))
            for line in extract_tb(tb):
                syslog(LOG_ERR, '  File "%s", line %i, in %s\n    %s\n' % (line))
            del t, v, tb
            
    def Run(self, nodelist):
        if self.element.get('state') not in ['ready', 'queued', 'stage-pending', 'prologue']:
            syslog(LOG_INFO, "Got multiple run commands for job %s" % self.attr['jobid'])
            return
        self.timers['queue'].Stop()
        syslog(LOG_INFO, "Running job %s on %s" % (self.attr['jobid'], ":".join(nodelist)))
        self.element.set('location', ":".join(nodelist))
        self.SetActive()

    def SetupNodes(self):
        self.attr['state'] = 'setup'
        self.SetActive()

    def CleanupNodes(self):
        self.attr['state'] = 'cleanup'
        self.SetActive()

    def FinishStage(self):
        self.attr['state'] = 'stage-pending'
        if not self.staged:
            self.SetPassive()
            self.steps = ['FinishStage'] + self.steps
        else:
            self.attr['state'] = 'ready'
            AddEvent("queue-manager", "job-ready", self.attr['jobid'])
            self.SetActive()     

    def FinishUserPgrp(self):
        self.timers['user'].Stop()
        if self.spgid.has_key('user'):
            try:
                h = comm.ClientInit("process-manager")
                comm.SendMessage(h, "<wait-process-group><process-group pgid='%s'><output/><error/></process-group></wait-process-group>" % self.spgid['user'])
                resp = comm.RecvMessage(h)
                comm.ClientClose(h)
                #if len(resp) > 8388608:
                #    syslog(LOG_ERR, "output is %d long for jobid %s, user %s; deleting"%(len(resp), self.attr['jobid'], self.attr['user']))
                #    r = Element("process-group")
                #else:
                self.output = ''
                self.error = ''
                r = XML(resp)
            except (CommunicationError, ConnectError, LookupError):
                syslog(LOG_ERR, "Error contacting process manager for finalize, requeueing")
                self.steps = ['FinalizeStage'] + self.steps
                self.SetActive()
                return
            except:
                (t, v, tb) = exc_info()
                syslog(LOG_ERR, "Error occurred while communicating with process-manager")
                for line in extract_tb(tb):
                    syslog(LOG_ERR, '  File "%s", line %i, in %s\n    %s\n' % line)
                    syslog(LOG_ERR, "%s: %s\n" % (t, v))
                del t, v, tb
                self.SetPassive()
                return

            for pg in r.findall("process-group"):
                od = pg.find("output")
                if od != None and od.text != None:
                    self.output = od.text
                else:
                    self.output = ""
                ed = pg.find("error")
                if ed != None and ed.text != None:
                    self.error = ed.text
                else:
                    self.error = ""
        else:
            syslog(LOG_ERR, "No record of pgid for user job %s" % (self.attr['jobid']))
        self.SetActive()

    def FinalizeStage(self):
        try:
            h = comm.ClientInit(self.scomp)
            if self.spgid.has_key('user'):
                comm.SendMessage(h,
                                 "<WriteOutput type='output' id='%s'>%s</WriteOutput>" % (self.stageid,
                                                                                          self.output))
                comm.RecvMessage(h)
                comm.SendMessage(h,
                                 "<WriteOutput type='error' id='%s'>%s</WriteOutput>" % (self.stageid,
                                                                                         self.error))
                comm.RecvMessage(h)
            comm.SendMessage(h, "<FinishStage id='%s'/>" % self.stageid)
            resp = comm.RecvMessage(h)
            comm.ClientClose(h)
        except:
            syslog(LOG_ERR, "Failed to contact %s for finalize, requeuing" % (self.scomp))
            self.steps = ['FinalizeStage'] + self.steps
        self.SetActive()

    def StageInit(self):
        # we need to assess which parts of stage need to complete
        msg = Element("SetupStage", outputdir=self.attr['outputdir'], name=self.attr['jobid'],
                      size=self.attr['nodes'], user=self.attr['user'])
        script = SubElement(msg, "script")
        script.text = escape(self.GetScript())
        if self.attr.get('stagein', None):
            msg.attrib['in'] = self.attr['stagein']
        if self.attr.get('stageout', None):
            msg.attrib['out'] = self.attr['stageout']

        try:
            h = comm.ClientInit("file-stager")
            comm.SendMessage(h, tostring(msg))
            r = comm.RecvMessage(h)
            comm.ClientClose(h)
            d = XML(r)
            if d.tag == 'error':
                syslog(LOG_ERR, "Stage System Error: %s" % (d.attrib['type']))
                self.attr['state'] = 'stage-error'
                self.steps = []
                return
            
            self.url = d.attrib['uri']
            self.stageid = d.attrib['id']
            self.scomp = d.attrib['component']
            syslog(LOG_DEBUG, "Got stageid %s for job %s" % (self.stageid, self.attr['jobid']))
        except LookupError:
            syslog(LOG_ERR, "Error connnecting to file-stager")
            self.attr['state'] = 'stage-error'
            self.steps = []
        except:
            (t, v, tb)=exc_info()
            syslog(LOG_ERR, "Error staging for job %s" % (self.attr['jobid']))
            for line in extract_tb(tb):
                syslog(LOG_ERR, '  File "%s", line %i, in %s\n    %s\n' % line)
                syslog(LOG_ERR, "%s: %s\n" % (t, v))
            del t, v, tb
            self.attr['state'] = 'stage-error'
            self.steps = []

    def RunPrologue(self):
        if self.element.get('location') == 'none':
            # requeue if not ready
            self.steps = ['RunPrologue'] + self.steps
            self.SetPassive()
            return
        self.attr['state'] = 'prologue'
        system("/master/bcfg/generators/account/setaccess.py -a %s %s" % (self.attr['user'],
                                                                          " ".join(self.element.get('location').split(':'))))
        self.timers['/usr/sbin/prologue'].Start()
        self.AdminStart('/usr/sbin/prologue')
        self.SetPassive()

    def RunEpilogue(self):
        self.attr['state'] ='epilogue'
        system("/master/bcfg/generators/account/setaccess.py -r %s %s" % (self.attr['user'],
                                                                          " ".join(self.element.get('location').split(':'))))
        self.timers['/usr/sbin/epilogue'].Start()
        self.AdminStart('/usr/sbin/epilogue')
        self.SetPassive()

    def RunUserJob(self):
        self.attr['state'] = 'running'
        self.timers['user'].Start()
        cpg = Element("create-process-group", submitter=self.attr['user'],
                      pgid='*', totalprocs=self.attr['nodes'])
        ps = Element("process-spec", path="/bin:/usr/bin:/usr/local/bin", cwd='/',
                     user=self.attr['user'])
        ps.attrib['exec'] = '/usr/bin/mpish'
        stageout = 0
        args = []
        if self.attr.get("host", None):
            args = ["-i", "-h", self.attr['host'], "-p", self.attr['port']]
        if self.attr.get("url", None):
            args += ["-b", self.url]
        if self.attr.get("stageid", None):
            args += ["-n", self.stageid]
        else:
            args += ["-c", self.GetScript()]
        if self.attr.get("stageout", None):
            args += ["-s", self.attr['stageout']]
        if self.attr.get("type", None) == 'pbs':
            args.append("-P")
        args.append("-t")
        args.append(str(60 * float(self.attr['walltime'])))
        idx = 1
        for arg in args:
            SubElement(ps, "arg", idx=str(idx), value=arg)
            idx += 1
        for e in self.element.findall("env"):
            SubElement(ps, "env", name=e.attrib['name'], value=e.attrib['value'])
        cpg.append(ps)
        hs = Element("host-spec")
        hs.text = "\n".join(self.element.get('location').split(':'))
        cpg.append(hs)
        try:
            h = comm.ClientInit("process-manager")
            comm.SendMessage(h, tostring(cpg))
            resp = comm.RecvMessage(h)
            comm.ClientClose(h)
        except (ConnectError, CommunicationError, ProtocolError, SocketError):
            raise ProcessManagerError

        try:
            r = XML(resp)
        except:
            raise ProcessManagerError

        self.pgid['user'] = r.attrib['pgid']
        self.SetPassive()

    def Kill(self, msg=None):
        if self.killed == True:
            return
        self.killed = True
        if msg != None:
            syslog(LOG_INFO, msg % (self.attr['jobid']))
        s = self.attr['state']
        j = self.attr['jobid']

        qreq = '''<del-lien id='%s'><project name='%s' /></del-lien>''' % (self.element.get('lienID'), self.element.get('project'))

        if s in ['epilogue', 'cleanup']:
            syslog(LOG_INFO, "Not killing job %s during recovery" % (j))
        elif s in ['setup', 'prologue']:
            # don't run job and just cleanup
            try:
                handle = comm.ClientInit("allocation-manager")
                comm.SendMessage(handle, qreq)
                r = comm.RecvMessage(handle)
            except:
                syslog(LOG_ERR, "Failed to delete lien id %s for project %s" % (self.element.get('lienID'), self.element.get('project')))
            self.steps.remove('RunUserJob')
        elif s == 'running':
            if not self.pgid.has_key('user'):
                syslog(LOG_ERR, "Job %s has no pgroup associated with it" % j)
            else:
                self.KillPGID(self.pgid['user'])
        elif s == 'stage-pending':
            self.steps = ['RunEpilogue', 'Finish']
            try:
                handle = comm.ClientInit("allocation-manager")
                comm.SendMessage(handle, qreq)
                r = comm.RecvMessage(handle)
            except:
                syslog(LOG_ERR, "Failed to delete lien id %s for project %s" % (self.element.get('lienID'), self.element.get('project')))
            self.SetActive()
        elif s == 'stage-error':
            try:
                handle = comm.ClientInit("allocation-manager")
                comm.SendMessage(handle, qreq)
                r = comm.RecvMessage(handle)
            except:
                syslog(LOG_ERR, "Failed to delete lien id %s for project %s" % (self.element.get('lienID'), self.element.get('project')))
            s = 'done'
            self.SetPassive()
        elif s == 'pm-error':
            try:
                handle = comm.ClientInit("allocation-manager")
                comm.SendMessage(handle, qreq)
                r = comm.RecvMessage(handle)
            except:
                syslog(LOG_ERR, "Failed to delete lien id %s for project %s" % (self.element.get('lienID'), self.element.get('project')))
            s = 'done'
            self.SetPassive()
        else:
            syslog(LOG_ERR, "Got qdel for job %s in unexpected state %s" % (j, s))

    def AdminStart(self, cmd):
        cpg = Element("create-process-group", submitter='root', pgid='*',
                      totalprocs=self.attr['nodes'], output='merged')
        ps = Element("process-spec", path="/bin:/usr/bin:/usr/local/bin", cwd='/', user='root')
        ps.attrib['exec'] = cmd
        cpg.append(ps)
        SubElement(ps, 'arg', idx='1', value=self.attr['user'])
        hs = Element("host-spec")
        hs.text = "\n".join(self.element.get('location').split(':'))
        ps.append(hs)
        
        try:
            h = comm.ClientInit('process-manager')
            comm.SendMessage(h, tostring(cpg))
            resp = comm.RecvMessage(h)
            comm.ClientClose(h)
            r = XML(resp)
        except:
            syslog(LOG_ERR, "Error running cmd %s for jobid %s"%(cmd, self.attr['jobid']))
            self.state = 'pm-error'
            return
        
        self.pgid[cmd] = r.attrib['pgid']

    def CompletePG(self, pgid):
        for (t, pg) in self.pgid.iteritems():
            if pg == pgid:
                syslog(LOG_INFO, "%s completed for job %s" % (t, self.attr['jobid']))
                self.spgid[t] = self.pgid[t]
                del self.pgid[t]
                self.timers[t].Stop()
                self.SetActive()
                return
        raise KeyError, pgid

    def KillPGID(self, pgid):
        try:
            h = comm.ClientInit("process-manager")
            comm.SendMessage(h,
                             "<kill-process-group><process-group pgid='%s'/></kill-process-group>" % pgid)
            resp = comm.RecvMessage(h)
            # need to process message
            comm.ClientClose(h)
        except (ConnectError, CommunicationError, SocketError, ProtocolError):
            syslog(LOG_ERR, "Error killing process group %s"%(pgid))
            raise ProcessManagerError

    # done to here

    def OverRun(self):
        if self.element.get('state') == 'running':
            runtime = self.timers['user'].Elapsed()/60.0
            if float(self.element.get('walltime')) < runtime:
                return 1
        return 0

#     # Here begins the testbed stuff
#     def StartRebuild(self):
#         self.Rebuild(self.attr['image'], self.attr['kernel'])

#     def CleanUpRebuild(self):
#         #self.Rebuild(defaultimage, defaultkernel)
#         pass

#     def Rebuild(self, image, kernel):
#         # send a message to the build system starting rebuild
#         self.built = 0
#         self.booted = 0
#         # send the message to the build system
#         h = comm.ClientInit("build-system")
#         i = comm.ClientInit("cluster-hardware")
#         for node in self.nodelist:
#             comm.SendMessage(h, "<set-node-software image='%s' kernel='%s' action='build'><node-software node='%s'/></set-node-software>" % (image, kernel, node))
#             comm.SendMessage(i, "<power-control type='reboot'><node-hardware node='%s'/></power-control>" % (node))
#             r = comm.RecvMessage(h)
#             r = comm.RecvMessage(i)
#         comm.ClientClose(h)
#         comm.ClientClose(i)
#         self.steps = ["BuildWait"] + self.steps
#         self.SetPassive()

#     def BuildWait(self):
#         if not ((self.built >= 2 * int(self.attr['nodes'])) and (self.booted >= 2 * int(self.attr['nodes']))):
#             self.steps = ["BuildWait"] + self.steps
#             self.SetPassive()
#         else:
#             self.SetActive()

    def GetStats(self):
        result = ''
        for (name, timer) in self.timers.iteritems():
            try:
                result += "%s:%.02fs " % (name, timer.Check())
            except:
                syslog(LOG_ERR, "Failed to check timer %s" % name)
        return result

    def LogFinish(self):
        syslog(LOG_INFO, "Job %s/%s on %s nodes done. %s" % (self.element.get('jobid'),
                                                             self.element.get('user'), self.element.get('nodes'), self.GetStats()))
        AddEvent("queue-manager", "job-done", self.attr['jobid'])

class BGJob(Job):
    '''BG Job is a Blue Gene/L job'''
    _configfields = ['bgkernel']
    _config = ConfigParser()
    if '-C' in argv:
        _config.read(argv[argv.index('-C') + 1])
    else:
        _config.read('/etc/cobalt.conf')
        if not _config._sections.has_key('cqm'):
            print '''"cqm" section missing from cobalt config file'''
            raise SystemExit, 1
    config = _config._sections['cqm']
    mfields = [field for field in _configfields if not config.has_key(field)]
    if mfields:
        print "Missing option(s) in cobalt config file: %s" % (" ".join(mfields))
        raise SystemExit, 1

    def __init__(self, element, jobid):
        self.element = element
        self.attr = self.element.attrib
        self.element.attrib['jobid'] = str(jobid)
        self.element.attrib['state'] = 'queued'
        self.element.set('location', 'N/A')
        if not self.element.get('queue', False):
            self.element.set('queue', 'default')
        if not self.element.get('kernel', False):
            self.element.set('kernel', 'default')
        self.pgid = {}
        self.spgid = {}
        self.killed = False
        self.timers = {'queue':Timer(), 'user':Timer()}
        AddEvent("queue-manager", "job-submitted", self.attr['jobid'])
        self.steps = ['RunBGUserJob', 'FinishUserPgrp', 'Finish']
        if self.config.get('bgkernel', 'false') == 'true':
            self.steps = ['SetBGKernel', 'RunBGUserJob', 'FinishUserPgrp', 'Finish']            
        self.timers['queue'].Start()
        self.SetPassive()

    def SetBGKernel(self):
        '''Ensure that the kernel is set properly prior to job launch'''
        current = readlink('/bgl/argonne-utils/partitions/%s' % (self.element.get('location')))
        if current != "/bgl/argonne-utils/profiles/%s" % self.element.get('kernel'):
            syslog(LOG_INFO, "Updating boot image for %s" % (self.element.get('location')))
            syslog(LOG_INFO, "Set to %s should be %s" % (current.split('/')[-1], self.element.get('kernel')))
            unlink('/bgl/argonne-utils/partitions/%s' % (self.element.get('location')))
            symlink('/bgl/argonne-utils/profiles/%s' % (self.element.get('kernel')),
                    '/bgl/argonne-utils/partitions/%s' % (self.element.get('location')))

    def RunBGUserJob(self):
        '''Run a Blue Gene Job'''
        self.element.attrib['state'] = 'running'
        self.timers['user'].Start()
        if not self.element.get('outputpath', False):
            self.element.set('outputpath', "%s/%s.output" % (self.element.get('outputdir'), self.attr['jobid']))
        if not self.element.get('errorpath', False):
            self.element.set('errorpath',
                             "%s/%s.error" % (self.element.get('outputdir'), self.element.get('jobid')))
        cpg = Element("create-process-group", submitter=self.attr['user'],
                      pgid='*', totalprocs=self.attr['procs'],
                      outputfile=self.element.get('outputpath'),
                      errorfile=self.element.get('errorpath'), mode=self.element.get('mode', 'co'))
        procspec = SubElement(cpg, "process-spec", path=self.attr['path'],
                        cwd=self.element.attrib['outputdir'], user=self.attr['user'])
        procspec.attrib['exec'] = self.element.find('command').get('bin')
        SubElement(procspec, 'arg', idx='1', value=self.element.find('command').get('args'))
        for env in self.element.findall("env"):
            SubElement(procspec, "env", name=env.attrib['name'], value=env.attrib['value'])
        hostspec = SubElement(cpg, "host-spec")
        hostspec.text = "\n".join(self.element.get('location').split(':'))
        try:
            handle = comm.ClientInit("process-manager")
            comm.SendMessage(handle, tostring(cpg))
            resp = comm.RecvMessage(handle)
            comm.ClientClose(handle)
        except (ConnectError, CommunicationError, ProtocolError, SocketError, SSSLookupError):
            raise ProcessManagerError

        try:
            response = XML(resp)
        except:
            raise ProcessManagerError

        if not response.attrib.has_key('pgid'):
            syslog(LOG_ERR, "Process Group creation failed for Job %s" % self.element.get('jobid'))
            self.element.set('state', 'pm-failure')
        else:
            self.pgid['user'] = response.attrib['pgid']
        self.SetPassive()
        
class CQM(EventReceiver):
    __implementation__ = 'cqm'
    __component__ = 'queue-manager'
    __statefields__ = ['Jobs']
    __subscriptions__ = [('process-manager', 'process_end', '**'),
                         ('node-state-manager', 'state-changed', '**'), ('fslave', '**', '**')]
    __dispatch__ = {'add-job':'Jobs.Add', 'get-job':'Jobs.Get', 'del-job':'XMLDelJob',
                    'run-job':'XRunJob', 'switch-queue':'XSwitchQueue', 'events':'HandleEvent'}
    __validate__ = 0

    def __setup__(self):
        self.Jobs = DataSet("jobs", "job", BGJob, ID(), False)

    def __progress__(self):
        [j.Progress() for j in self.Jobs if j.active]
        [j.Kill("Job %s Overtime, Killing") for j in self.Jobs if j.OverRun()]
        [j.LogFinish() for j in self.Jobs if j.attr['state'] == 'done']
        [self.Jobs.remove(j) for j in self.Jobs if j.attr['state'] == 'done']
        return 1

    def __postbind__(self):
        signal(SIGCHLD, SIG_DFL)
        EventReceiver.__postbind__(self)

    def XRunJob(self, xml, (peer, port)):
        '''Process the runjob callback'''
        return self.Jobs.Get(xml, (peer, port), self.RunJob)

    def RunJob(self, job, attr):
        '''Run the job'''
        job.Run(attr['nodelist'].split(':'))

    def XSwitchQueue(self, xml, portinfo):
        '''Handle queue change'''
        return self.Jobs.Get(xml, portinfo, self.SwitchQueue)

    def SwitchQueue(self, job, attr):
        '''Change the queue for job'''
        job.element.set("queue", attr['queue'])

    def XMLDelJob(self, xml, (peer, port)):
        r = Element("jobs")
        for spec in xml.findall('job'):
            [r.append(job.Fetch(spec)) for job in self.Jobs if job.Match(spec)]
            for job in [job for job in self.Jobs if job.Match(spec)]:
                if ((job.attr['state'] == 'queued') or (job.attr['state'] == 'ready') or (xml.attrib.has_key('force'))):
                    self.Jobs.remove(job)
                else:
                    job.Kill()
                AddEvent("queue-manager", "job-deleted", job.attr['jobid'])
        return r

    def HandleEvent(self, xml, portinfo):
        for event in xml.findall('event'):
            (c, m, d) = tuple([event.get(field) for field in ['component', 'msg', 'data']])
            if c == 'node-stage-manager':
                # jobs gate on nsm up message for all nodes in job
                j = [j for j in self.Jobs if d in j.nodelist]
                if len(j) == 1:
                    j[0].booted += 1                    
                else:
                    pass
            elif c == 'build-system':
                # jobs gate on bs state change for action = boot set
                j = [j for j in self.Jobs if d in j.nodelist]
                if len(j) == 1:
                    j[0].built += 1                    
                else:
                    pass
            elif c == 'fslave':
                # d is the stageid
                jl = [j for j in self.Jobs if j.stageid == d]
                if len(jl) == 0:
                    syslog(LOG_INFO, "Got unknown stage id %s" % (d))
                elif len(jl) == 1:
                    if m == 'stage-setup':
                        syslog(LOG_DEBUG,
                               "Got stage completed for stageid %s, jobid %s" % (d, jl[0].attr['jobid']))
                        if jl[0].attr['state'] == 'stage-pending':
                            jl[0].SetActive()
                        AddEvent("queue-manager", "job-ready", str(jl[0].attr['jobid']))
                        jl[0].staged = 1
                    elif m == 'stage-failed':
                        jl[0].attr['state'] = 'stage-error'
                    else:
                        syslog(LOG_ERR, "Unexpected fslave event")
                else:
                    syslog(LOG_ERR, "Got multiple matches for stageid %s" % (d))
            elif c == 'process-manager':
                # Something has exited
                [j.CompletePG(d) for j in self.Jobs if j.HasPG(d)]
        return Element('event-ok')

    def PMResync(self):
        try:
            h = comm.ClientInit("process-manager")
            comm.SendMessage(h,
                             '''<get-process-group-info><process-group pgid='*'/></get-process-group-info>''')
            r = comm.RecvMessage(h)
            comm.ClientClose(h)
            msg = XML(r)
        except:
            syslog(LOG_ERR, "Error syncing with process-manager")
            return
        pgids = [pgrp.get('pgid') for pgrp in msg.findall("process-group")]
        for job in self.Jobs:
            for (t, pg) in job.pgid:
                if pg not in pgids:
                    syslog(LOG_INFO, "Found dead pg for job %s" % (job.attr['jobid']))
                    job.CompletePG(pg)
            
if __name__ == '__main__':
    from getopt import getopt, GetoptError
    try:
        (opts, args) = getopt(argv[1:], 'dC:', ['daemon='])
    except GetoptError, msg:
        print "%s\nUsage:\ncqm.py [-d] [-C <configfile>] [--daemon <pidfile>]" % (msg)
        raise SystemExit, 1
    daemon = [x[1] for x in opts if x[0] == '--daemon']
    debug = len([x for x in opts if x[0] == '-d'])
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0])
    server = CQM(debug=debug)
    server.ServeForever()
