#!/usr/bin/env python
# $Id: qm2.py 1.4 05/06/17 16:11:31-05:00 bradshaw@skywalker.mcs.anl.gov $

'''Queue Manager, Version 2'''
__revision__ = '$Revision: 1.4 $'

from elementtree.ElementTree import Element, SubElement, XML, tostring
from xml.sax.saxutils import escape
from os import system
from signal import signal, SIGCHLD, SIG_DFL
from sys import exc_info
from syslog import syslog, LOG_INFO, LOG_ERR, LOG_DEBUG
from time import time
from traceback import extract_tb
#from smtplib import SMTP

from sss.restriction import Data, DataSet, ID
from sss.server import EventReceiver, AddEvent
from sss.ssslib import comm_lib, ConnectError, CommunicationError, ProtocolError, SocketError, LookupError as SSSLookupError

class ProcessManagerError(Exception):
    '''This error occurs when communications with the process manager fail'''
    pass

class TimerException(Exception):
    '''This error occurs when timer methods are called in the wrong order'''
    pass

class Timer(object):
    def __init__(self):
        pass
    
    def Start(self):
        self.start = time()
        
    def Stop(self):
        self.stop = time()
        
    def Check(self):
        try:
            return self.stop - self.start
        except:
            raise TimerException, "StateError"

    def Elapsed(self):
        try:
            return time() - self.start
        except:
            raise TimerException, "StateError"

comm = comm_lib(debug=0)

class Job(Data):
    '''The Job class is an object corresponding to the qm notion of a queued job, including steps'''
    def __init__(self, e, jobid):
        self.element = e
        self.attr = self.element.attrib
        self.attr['jobid'] = str(jobid)
        self.attr['state'] = 'queued'
        self.attr['attribute'] = 'compute'
        self.staged = 0
        self.killed = False
        self.output = ''
        self.error = ''
        self.timers = {}
        self.timers['queue'] = Timer()
        self.timers['queue'].Start()
        self.timers['/usr/sbin/prologue'] = Timer()
        self.timers['user'] = Timer()
        self.timers['/usr/sbin/epilogue'] = Timer()
        self.pgid = {}
        self.spgid = {}
        self.steps = ['CheckProject', 'StageInit', 'FinishStage', 'RunPrologue',
                      'RunUserJob', 'FinishUserPgrp', 'RunEpilogue', 'FinalizeStage', 'Finish']
        #self.steps=['StageInit','FinishStage','RunPrologue','RunUserJob','RunEpilogue','FinalizeStage','Finish']
        self.nodelist = []
        self.stageid = None
        if not self.attr.get('type', None):
            self.attr['type'] = 'mpish'
        AddEvent("queue-manager", "job-submitted", self.attr['jobid'])
        self.SetActive()

    def CheckProject(self):
        global comm
        try:
            req = '''<get-project><project name='%s'><user name='%s'/></project></get-project>''' % (
                self.attr['project'], self.attr['user'])
            h = comm.ClientInit('allocation-manager')
            comm.SendMessage(h, req)
            r = comm.RecvMessage(h)
            comm.ClientClose(h)
        except:
            syslog(LOG_ERR, "Failed to contact allocation manager")
            print "failed to contact AM"
            return
        
        if not len(XML(r).getchildren()):
            # they don't have a project
            syslog(LOG_ERR, "Allocation error for user %s submitting with project %s (project doesn't exist)"%(self.attr['user'], self.attr['project']))
            self.attr['state'] = 'am-error'
            self.steps = []
            return
        else:
            try:
                #add-lien is specified in time of seconds
                req = '''<add-lien amount='%s'><project name='%s'/></add-lien>''' % (
                    str(float((self.attr['walltime'])) * 60), self.attr['project'])

                h = comm.ClientInit('allocation-manager')
                comm.SendMessage(h, req)
                r = comm.RecvMessage(h)
                comm.ClientClose(h)
            except:
                syslog(LOG_ERR, "Failed to contact allocation manager")
                self.attr['state'] = 'am-error'
                self.steps = []
                return

            try:
                self.lienID = XML(r).find('lien').get('id')
            except:
                syslog(LOG_ERR, "Allocation error for user %s submitting with project %s (insufficient balance)"%(self.attr['user'], self.attr['project']))
                self.attr['state'] = 'am-error'
                self.steps = []
                return
        return

    def SetActive(self):
        self.active = True

    def SetPassive(self):
        self.active = False
        
    def HasPG(self, pgid):
        if pgid in self.pgid.values():
            return 1
        return 0

    def Finish(self):
        #am-stuff
        time = int(self.timers['user'].Check()) * len(self.nodelist)
        try:
            req = '''<commit-lien id='%s' actual='%s'><project name='%s'/></commit-lien>''' % (str(self.lienID), str(time), self.attr['project'])
            h = comm.ClientInit('allocation-manager')
            comm.SendMessage(h, req)
            r = comm.RecvMessage(h)
            comm.ClientClose(h)
        except:
            syslog(LOG_ERR, "Failed to contact allocation manager")
                                                                
        self.attr['state'] = 'done'
        self.SetPassive()
        AddEvent("queue-manager", "job-completed", self.attr['jobid'])

    def GetScript(self):
        return self.element.find('script').text

    def Progress(self):
        if not self.steps:
            syslog(LOG_ERR, "Manually setting passive for job %s" % (self.attr['jobid']))
            self.SetPassive()
            return
        syslog(LOG_INFO, "running step %s for job %s" % (self.steps[0], self.attr['jobid']))
        getattr(self, self.steps[0])()
        try:
            self.steps = self.steps[1:]
        except IndexError:
            syslog(LOG_ERR, "No more job steps for job %s" % (self.attr['jobid']))
            self.attr['state'] = 'error'
        except:
            syslog(LOG_ERR, "Unexpected failure jobid:%s step:%s" % (self.attr['jobid'], self.steps[0]))
            (t, v, tb) = exc_info()
            syslog(LOG_ERR, "Error: %s: %s" % (t, v))
            for line in extract_tb(tb):
                syslog(LOG_ERR, '  File "%s", line %i, in %s\n    %s\n' % (line))
            del t, v, tb
            
    def Run(self, nodelist):
        if self.element.get('state') not in ['queued','stage-pending', 'prologue', 'ready']:
            syslog(LOG_INFO, "Got multiple run commands for job %s" % self.attr['jobid'])
            return
        self.timers['queue'].Stop()
        syslog(LOG_INFO, "Running job %s on %s" % (self.attr['jobid'], ":".join(nodelist)))
        self.nodelist = nodelist
        self.SetActive()

    def SetupNodes(self):
        self.attr['state'] = 'setup'
        self.SetActive()

    def CleanupNodes(self):
        self.attr['state'] = 'cleanup'
        self.SetActive()

    def FinishStage(self):
        self.attr['state'] = 'stage-pending'
        if not self.staged:
            self.SetPassive()
            self.steps = ['FinishStage'] + self.steps
        else:
            self.attr['state'] = 'ready'
            AddEvent("queue-manager", "job-ready", self.attr['jobid'])
            self.SetActive()     

    def FinishUserPgrp(self):
        self.timers['user'].Stop()
        if self.spgid.has_key('user'):
            try:
                h = comm.ClientInit("process-manager")
                comm.SendMessage(h, "<wait-process-group><process-group pgid='%s'><output/><error/></process-group></wait-process-group>" % self.spgid['user'])
                resp = comm.RecvMessage(h)
                comm.ClientClose(h)
                #if len(resp) > 8388608:
                #    syslog(LOG_ERR, "output is %d long for jobid %s, user %s; deleting"%(len(resp), self.attr['jobid'], self.attr['user']))
                #    r = Element("process-group")
                #else:
                r = XML(resp)
            except (CommunicationError, ConnectError, SSSLookupError):
                syslog(LOG_ERR, "Error contacting process manager for finalize, requeueing")
                self.steps = ['FinishUserPgrp'] + self.steps
                self.SetActive()
                return
            except:
                (t, v, tb) = exc_info()
                syslog(LOG_ERR, "Error occurred while communication with process-manager")
                for line in extract_tb(tb):
                    syslog(LOG_ERR, '  File "%s", line %i, in %s\n    %s\n' % line)
                    syslog(LOG_ERR, "%s: %s\n" % (t, v))
                del t, v, tb
                self.SetPassive()
                return

            for pg in r.findall("process-group"):
                od = pg.find("output")
                if od != None and od.text != None:
                    self.output = od.text
                ed = pg.find("error")
                if ed != None and ed.text != None:
                    self.error = ed.text
        else:
            syslog(LOG_ERR, "No record of pgid for user job %s" % (self.attr['jobid']))
        self.SetActive()

    def FinalizeStage(self):
        try:
            h = comm.ClientInit(self.scomp)
            if self.spgid.has_key('user'):
                comm.SendMessage(h,
                                 "<WriteOutput type='output' id='%s'>%s</WriteOutput>" % (
                    self.stageid, self.output))
                comm.RecvMessage(h)
                comm.SendMessage(h,
                                 "<WriteOutput type='error' id='%s'>%s</WriteOutput>" % (
                    self.stageid, self.error))
                comm.RecvMessage(h)
            comm.SendMessage(h, "<FinishStage id='%s'/>" % self.stageid)
            resp = comm.RecvMessage(h)
            comm.ClientClose(h)
        except:
            syslog(LOG_ERR, "Failed to contact %s for finalize, requeuing" % (self.scomp))
            self.steps = ['FinalizeStage'] + self.steps
        self.SetActive()

    def StageInit(self):
        # we need to assess which parts of stage need to complete
        msg = Element("SetupStage", outputdir=self.attr['outputdir'], name=self.attr['jobid'],
                      size=self.attr['nodes'], user=self.attr['user'])
        script = SubElement(msg, "script")
        script.text = escape(self.GetScript())
        if self.attr.get('stagein', None):
            msg.attrib['in'] = self.attr['stagein']
        if self.attr.get('stageout', None):
            msg.attrib['out'] = self.attr['stageout']

        try:
            h = comm.ClientInit("file-stager")
            comm.SendMessage(h, tostring(msg))
            r = comm.RecvMessage(h)
            comm.ClientClose(h)
            d = XML(r)
            if d.tag == 'error':
                syslog(LOG_ERR, "Stage System Error: %s" % (d.attrib['type']))
                self.attr['state'] = 'stage-error'
                self.steps = []
                return
            
            self.url = d.attrib['uri']
            self.stageid = d.attrib['id']
            self.scomp = d.attrib['component']
            syslog(LOG_DEBUG, "Got stageid %s for job %s" % (self.stageid, self.attr['jobid']))
        except LookupError:
            syslog(LOG_ERR, "Error connnecting to file-stager")
            self.attr['state'] = 'stage-error'
            self.steps = []
        except:
            (t, v, tb)=exc_info()
            syslog(LOG_ERR, "Error staging for job %s" % (self.attr['jobid']))
            for line in extract_tb(tb):
                syslog(LOG_ERR, '  File "%s", line %i, in %s\n    %s\n' % line)
                syslog(LOG_ERR, "%s: %s\n" % (t, v))
            del t, v, tb
            self.attr['state'] = 'stage-error'
            self.steps = []

    def RunPrologue(self):
        if not self.nodelist:
            self.steps = ['RunPrologue'] + self.steps
            self.SetPassive()
            return
        self.attr['state'] = 'prologue'
        #system("/master/bcfg/generators/account/setaccess.py -a %s %s" % (self.attr['user'],
        #                                                                  " ".join(self.nodelist)))
        self.timers['/usr/sbin/prologue'].Start()
        self.AdminStart('/usr/sbin/prologue')
        self.SetPassive()

    def RunEpilogue(self):
        self.attr['state'] ='epilogue'
        #system("/master/bcfg/generators/account/setaccess.py -r %s %s" % (self.attr['user'],
        #                                                                  " ".join(self.nodelist)))
        self.timers['/usr/sbin/epilogue'].Start()
        self.AdminStart('/usr/sbin/epilogue')
        self.SetPassive()

    def RunUserJob(self):
        self.attr['state'] = 'running'
        self.timers['user'].Start()
        cpg = Element("create-process-group", submitter=self.attr['user'],
                      pgid='*', totalprocs=self.attr['procs'], output="merged")
        ps = Element("process-spec", path="/bin:/usr/bin:/usr/local/bin", cwd='/',
                     user=self.attr['user'])
        ps.attrib['exec'] = '/usr/bin/mpish'
        stageout = 0
        pargs = []
        if self.attr.get("host", None):
            pargs = ["-i", "-h", self.attr['host'], "-p", self.attr['port']]
        pargs += ["-b", self.url, "-n", self.stageid]
        if self.attr.get("stageout", None):
            pargs += ["-s", self.attr['stageout']]
        if self.attr.get("type", None) == 'pbs':
            pargs.append("-P")
        pargs.append("-t")
        pargs.append(str(60 * float(self.attr['walltime'])))
        idx = 1
        for arg in pargs:
            SubElement(ps, "arg", idx=str(idx), value=arg)
            idx += 1
        for e in self.element.findall("env"):
            SubElement(ps, "env", name=e.attrib['name'], value=e.attrib['value'])
        cpg.append(ps)
        hs = Element("host-spec")
        hs.text = "\n".join(self.nodelist)
        cpg.append(hs)
        try:
            h = comm.ClientInit("process-manager")
            comm.SendMessage(h, tostring(cpg))
            resp = comm.RecvMessage(h)
            comm.ClientClose(h)
        except (ConnectError, CommunicationError, ProtocolError, SocketError):
            raise ProcessManagerError

        try:
            r = XML(resp)
        except:
            raise ProcessManagerError

        self.pgid['user'] = r.attrib['pgid']
        self.SetPassive()

    def Kill(self, msg=None):
        if self.killed == True:
            return
        self.killed = True
        if msg != None:
            syslog(LOG_INFO, msg % (self.attr['jobid']))
        s = self.attr['state']
        j = self.attr['jobid']
        if s in ['epilogue', 'cleanup']:
            syslog(LOG_INFO, "Not killing job %s during recovery" % (j))
        elif s in ['setup', 'prologue']:
            # don't run job and just cleanup
            self.steps.remove('RunUserJob')
        elif s == 'running':
            self.KillPGID(self.pgid['user'])
        elif s == 'stage-pending':
            self.steps = ['RunEpilogue', 'Finish']
            self.SetActive()
        elif s == 'stage-error':
            # no cleanup needed
            s = 'done'
            self.SetPassive()
        elif s == 'am-error':
            # no cleanup needed
            s = 'done'
            self.SetPassive()
        elif s == 'pm-error':
            # no cleanup needed
            s = 'done'
            self.SetPassive()
        else:
            syslog(LOG_ERR, "Got qdel for job %s in unexpected state %s" % (j, s))

    def AdminStart(self, cmd):
        cpg = Element("create-process-group", submitter='root', pgid='*',
                      totalprocs=self.attr['nodes'], output='merged')
        ps = Element("process-spec", path="/bin:/usr/bin:/usr/local/bin", cwd='/', user='root')
        ps.attrib['exec'] = cmd
        cpg.append(ps)
        SubElement(ps, 'arg', idx='1', value=self.attr['user'])
        hs = Element("host-spec")
        hs.text = "\n".join(self.nodelist)
        ps.append(hs)
        
        try:
            h = comm.ClientInit('process-manager')
            comm.SendMessage(h, tostring(cpg))
            resp = comm.RecvMessage(h)
            comm.ClientClose(h)
            r = XML(resp)
        except:
            syslog(LOG_ERR, "Error running cmd %s for jobid %s"%(cmd, self.attr['jobid']))
            self.state = 'pm-error'
            return
        
        self.pgid[cmd] = r.attrib['pgid']

    def AdminFinish(self, cmd):
        wpg = Element("wait-process-group", pgid=self.pgid[cmd])
        
        try:
            h = comm.ClientInit('process-manager')
            comm.SendMessage(h, tostring(wpg))
            resp = comm.RecvMessage(h)
            comm.ClientClose(h)
            r = XML(resp)
        except:
            syslog(LOG_ERR, "Error waiting on cmd %s with pgid %s for jobid %s"%(cmd, self.pgid[cmd], self.attr['jobid']))
            self.state = 'pm-error'
            return
        
    def CompletePG(self, pgid):
        for (t, pg) in self.pgid.iteritems():
            if pg == pgid:
                syslog(LOG_INFO, "%s completed for job %s" % (t, self.attr['jobid']))
                if t in ['/usr/sbin/prologue','/usr/sbin/epilogue' ]:
                    self.AdminFinish(t)
                self.spgid[t] = self.pgid[t]
                del self.pgid[t]
                self.timers[t].Stop()
                self.SetActive()
                return
        raise KeyError, pgid

    def KillPGID(self, pgid):
        try:
            h = comm.ClientInit("process-manager")
            comm.SendMessage(h,"<kill-process-group><process-group pgid='%s'/></kill-process-group>"%pgid)
            resp = comm.RecvMessage(h)
            # need to process message
            comm.ClientClose(h)
        except (ConnectError, CommunicationError, SocketError, ProtocolError):
            syslog(LOG_ERR, "Error killing process group %s"%(pgid))
            raise ProcessManagerError

    # done to here

    def OverRun(self):
        if self.element.get('state') == 'running':
            runtime = self.timers['user'].Elapsed()/60.0
            if float(self.element.get('walltime')) < runtime:
                return 1
        return 0

    # Here begins the testbed stuff
    def StartRebuild(self):
        self.Rebuild(self.attr['image'], self.attr['kernel'])

    def CleanUpRebuild(self):
        self.Rebuild(defaultimage, defaultkernel)

    def Rebuild(self, image, kernel):
        # send a message to the build system starting rebuild
        self.built = 0
        self.booted = 0
        # send the message to the build system
        h = comm.ClientInit("build-system")
        i = comm.ClientInit("cluster-hardware")
        for node in self.nodelist:
            comm.SendMessage(h, "<set-node-software image='%s' kernel='%s' action='build'><node-software node='%s'/></set-node-software>" % (image, kernel, node))
            comm.SendMessage(i, "<power-control type='reboot'><node-hardware node='%s'/></power-control>" % (node))
            r = comm.RecvMessage(h)
            r = comm.RecvMessage(i)
        comm.ClientClose(h)
        comm.ClientClose(i)
        self.steps = ["BuildWait"] + self.steps
        self.SetPassive()

    def BuildWait(self):
        if not ((self.built >= 2 * int(self.attr['nodes'])) and (self.booted >= 2 * int(self.attr['nodes']))):
            self.steps = ["BuildWait"] + self.steps
            self.SetPassive()
        else:
            self.SetActive()

    def GetStats(self):
        result = ''
        for (name, timer) in self.timers.iteritems():
            try:
                result += "%s:%.02fs " % (name, timer.Check())
            except:
                syslog(LOG_ERR, "Failed to check timer %s" % name)
        return result

    def LogFinish(self):
        syslog(LOG_INFO, "Job %s done. %s" % (self.element.get('jobid'), self.GetStats()))
        AddEvent("queue-manager", "job-done", self.attr['jobid'])

class QM(EventReceiver):
    __implementation__ = 'qm2'
    __component__ = 'queue-manager'
    __statefields__ = ['Jobs']
    __subscriptions__ = [('process-manager', 'process_end', '**'),
                         ('node-state-manager', 'state-changed', '**'), ('fslave', '**', '**')]
    __dispatch__ = {'add-job':'Jobs.Add', 'get-job':'Jobs.Get', 'del-job':'XMLDelJob',
                    'run-job':'XRunJob', 'events':'HandleEvent'}
    __validate__ = 0

    def __setup__(self):
        self.Jobs = DataSet("jobs", "job", Job, ID(), False)

    def __progress__(self):
        [j.Progress() for j in self.Jobs if j.active]
        [j.Kill("Job %s Overtime, Killing") for j in self.Jobs if j.OverRun()]
        [j.LogFinish() for j in self.Jobs if j.attr['state'] == 'done']
        [self.Jobs.remove(j) for j in self.Jobs if j.attr['state'] == 'done']
        return 1

    def __postbind__(self):
        signal(SIGCHLD, SIG_DFL)
        EventReceiver.__postbind__(self)

    def XRunJob(self, xml, (peer, port)):
        return self.Jobs.Get(xml, (peer, port), self.RunJob)

    def RunJob(self, job, attr):
        job.Run(attr['nodelist'].split(':'))

    def XMLDelJob(self, xml, (peer, port)):
        r = Element("jobs")
        for spec in xml.findall('job'):
            [r.append(job.Fetch(spec)) for job in self.Jobs if job.Match(spec)]
            for job in [job for job in self.Jobs if job.Match(spec)]:
                if ((job.attr['state'] == 'queued') or (job.attr['state'] == 'ready') or (xml.attrib.has_key('force'))):
                    self.Jobs.remove(job)
                else:
                    job.Kill()
                AddEvent("queue-manager", "job-deleted", job.attr['jobid'])
        return r

    def HandleEvent(self, xml, portinfo):
        for event in xml.findall('event'):
            (c, m, d) = tuple([event.get(field) for field in ['component', 'msg', 'data']])
            if c == 'node-stage-manager':
                # jobs gate on nsm up message for all nodes in job
                j = [j for j in self.Jobs if d in j.nodelist]
                if len(j) == 1:
                    j[0].booted += 1                    
                else:
                    pass
            elif c == 'build-system':
                # jobs gate on bs state change for action = boot set
                j = [j for j in self.Jobs if d in j.nodelist]
                if len(j) == 1:
                    j[0].built += 1                    
                else:
                    pass
            elif c == 'fslave':
                # d is the stageid
                jl = [j for j in self.Jobs if j.stageid == d]
                if len(jl) == 0:
                    syslog(LOG_INFO, "Got unknown stage id %s" % (d))
                elif len(jl) == 1:
                    if m == 'stage-setup':
                        syslog(LOG_DEBUG,
                               "Got stage completed for stageid %s, jobid %s" % (d, jl[0].attr['jobid']))
                        if jl[0].attr['state'] == 'stage-pending':
                            jl[0].SetActive()
                        AddEvent("queue-manager", "job-ready", str(jl[0].attr['jobid']))
                        jl[0].staged = 1
                    elif m == 'stage-failed':
                        jl[0].attr['state'] = 'stage-error'
                    else:
                        syslog(LOG_ERR, "Unexpected fslave event")
                else:
                    syslog(LOG_ERR, "Got multiple matches for stageid %s" % (d))
            elif c == 'process-manager':
                # Something has exited
                [j.CompletePG(d) for j in self.Jobs if j.HasPG(d)]
        return Element('event-ok')

    def PMResync(self):
        try:
            h = comm.ClientInit("process-manager")
            comm.SendMessage(h, '''<get-process-group-info><process-group pgid='*'/></get-process-group-info>''')
            r = comm.RecvMessage(h)
            comm.ClientClose(h)
            msg = XML(r)
        except:
            syslog(LOG_ERR, "Error syncing with process-manager")
            return
        pgids = [pgrp.get('pgid') for pgrp in msg.findall("process-group")]
        for job in self.Jobs:
            for (t, pg) in job.pgid:
                if pg not in pgids:
                    syslog(LOG_INFO, "Found dead pg for job %s" % (job.attr['jobid']))
                    job.CompletePG(pg)
            
if __name__ == '__main__':
    from getopt import getopt, GetoptError
    from sys import argv
    try:
        (opts, args) = getopt(argv[1:], 'd', ['daemon='])
    except GetoptError, gmsg:
        print "%s\nUsage:\nqm.py [-d] [--daemon <pidfile>]" % (gmsg)
        raise SystemExit, 1
    daemon = [x[1] for x in opts if x[0] == '--daemon']
    debug = len([x for x in opts if x[0] == '-d'])
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0])
    server = QM(debug=debug)
    server.ServeForever()
